/** 
 * @file Gainer.cpp
 * @brief Gainer Class
 * @author dandelion
 * @date 2009-04-18
 * @version $Id: Gainer.cpp, v0.1$
 *
 * Copyright (C) 2009 dandelion. All rights reserved.
 */


#include <iostream>
#include <string>
#include <sstream>

// Gainer class header
#include "Gainer.h"

// ERSLIB(http://www.geocities.jp/in_subaru/erslib/)
#include "erslib.h"

#ifdef DEBUG
#define DEBUG_PRINT(msg) std::cerr << msg << std::endl;
#else
#define DEBUG_PRINT(msg) ;
#endif // DEBUG

/**
 * Gainer class default constructor. COM port number is COM1.
 * @param portNum COM port number
 * @param mode configutation mode (1-7)
 * @param verbose verbose output
 */
Gainer::Gainer(int portNum, int mode, bool verbose) :
led_(false)
, portNum_(portNum)
, config_(mode)
, endFlag(false)
, on_pressed(NULL)
, on_released(NULL)
{
	DEBUG_PRINT("Gainer()");

	// COM port open
	ERS_Open(this->portNum_, 38400, 38400);

	// setting of COM port
	if(setupCOMPort() != 0)
	{
		DEBUG_PRINT("Error");
	}

	// setting receiving timeout(=100ms)
	ERS_RecvTimeOut(this->portNum_, 100);

	// create receiving thread
	this->handle = (HANDLE)_beginthreadex(NULL, 0, receiver, this, 0, NULL);

	// software reset
	reboot();

	// set configulation mode
	setConfiguration(config_);
}

/**
 * Gainer class destructor.
 */
Gainer::~Gainer()
{
	DEBUG_PRINT("~Gainer()");

	// software reset
	reboot();

	endFlag = true;
	WaitForSingleObject(this->handle, WAIT_TIMEOUT);
	CloseHandle(this->handle);

	// COM port close
	ERS_Close(this->portNum_);
}

/**
 * setup COM port.
 * @return 0: normal termination, 1: error, 2: error
 */
int Gainer::setupCOMPort()
{
	// setting of Serial communication parameter
	// (38400bps, non parity, stop bit=1bit, data bit=8bit)
	return ERS_Config(this->portNum_, ERS_38400 | ERS_NO | ERS_1 | ERS_8);
}

void Gainer::setConfiguration(int mode)
{
	// Configulation Mode check
	if (1 > mode || mode > 7)
	{
		throw GainerInternal::Exception();
	}

	config_ = mode;
	std::stringstream ss("");
	ss << "KONFIGURATION_" << mode;
	command(ss.str(), SEND_TIME);

	std::cout << "KONFIGURATION_" << mode << std::endl;

	for (int i(0); i<CONFIG[config_][AIN]; i++) // create analog inputs
	{
		analogInputs.push_back(0);
	}

	for (int i(0); i<CONFIG[config_][DIN]; i++) // create digital inputs
	{
		digitalInputs.push_back(false);
	}
}

/**
 * Turn on the LED on gainer. command "h*" 
 * @return
 */
void Gainer::turnOnLED()
{
	command("h", 1);
}

/**
 * Turn off the LED on gainer. command "l*" 
 * @return
 */
void Gainer::turnOffLED()
{
	command("l", 1);
}


void Gainer::peekDigitalInput()
{
	command("R");
}

void Gainer::peekAnalogInput()
{
	command("I");
}

// Not implemented yet. (2009/04/18) -------
#if 0
void Gainer::set_matrix()
{

}

void Gainer::continuous_digital_inputs()
{
	command("r");
}

void Gainer::continuous_analog_inputs()
{
	command("i");
}

void Gainer::exit_continuos()
{
	command("E");
}
#endif
//------------------------------------------

/**
 * It makes 0V or +5V on digital output port.
 * @param value values of the digital output ports true:1 false:0	
 * @return
 */
void Gainer::digitalOutput(int value)
{
	std::stringstream ss("");
	ss << "D";
	for (int i(0); i< CONFIG[config_][DOUT]-1; i++)
	{
		ss << "0";
	}
	ss << value;
	command(ss.str(), SEND_TIME);
}

/**
 * It makes +5V on digital output port.
 * @param port number of the digital output port
 * @return
 */
void Gainer::setHigh(int port)
{
	std::stringstream ss("");
	ss << "H" << port;
	command(ss.str(), SEND_TIME);
}

// Not implemented yet. (2009/04/18) -------
#if 0
/**
 * It makes +5V on digital output port.
 * @param portArray number of the digital output port
 * @return
 */
void Gainer::setHigh(int *portArray)
{

}
#endif
//------------------------------------------

/**
 * It makes 0V on digital output port.
 * @param port number of the digital output port
 * @return
 */
void Gainer::setLow(int port)
{
	std::stringstream ss("");
	ss << "L" << port;
	command(ss.str(), SEND_TIME);
}

// Not implemented yet. (2009/04/18) -------
#if 0
/**
 * It makes 0V on digital output port.
 * @param portArray number of the digital output port
 * @return
 */
void Gainer::setLow(int *portArray)
{

}
#endif
//------------------------------------------

/**
 * Software reset. User should not use this.
 * @return
 */
void Gainer::reboot()
{
	command("Q", SEND_TIME);
}

/**
 * 
 * @param cmd command string
 * @param wait wait time(ms)
 * @return
 */
void Gainer::command(const std::string &cmd, int wait)
{
	command_send(cmd + '*');
	::Sleep(wait);
}

void Gainer::command_send(const std::string &cmd)
{
	std::cout << "command_send : " << cmd.c_str() << std::endl;
	ERS_Puts(this->portNum_, const_cast<char *>(cmd.c_str()));
}

void Gainer::processEvent(std::string &event)
{
	DEBUG_PRINT("event: " << event);
	switch(event[0])
	{
	case '!': // something wrong
		throw GainerInternal::Exception();
	case 'h': // led on
		led_ = true;
		break;
	case 'l': // led off
		led_ = false;
		break;
	case 'N': // button pressed
		if(on_pressed)
		{
			on_pressed();
		}
		break;
	case 'F': // button released
		if(on_released)
		{
			on_released();
		}
		break;
	case 'i':
	case 'I': { // analog_input
		std::string::size_type ast(event.find('*'));
		std::string s(event.substr(1, ast-1));
		sscanf(s.c_str(), "%02X%02X%02X%02X*",
			&analogInputs[0], &analogInputs[1],
			&analogInputs[2], &analogInputs[3]);
		break;
			  }
	case 'r':
	case 'R': { // digital input
		std::string::size_type ast(event.find('*'));
		std::stringstream ss(event.substr(1, ast-1));
		int num;
		ss >> std::hex >> num;
		for (size_t i(0); i< digitalInputs.size(); i++)
		{
			digitalInputs[i] = (num & ( 1 << i));
		}
		break;
			  }
	default:
		break;
	}
}

unsigned __stdcall Gainer::receiver(void *arg)
{
	Gainer *self(reinterpret_cast<Gainer *>(arg));
	while (!self->endFlag)
	{
		const size_t SIZE(80);
		char buf[SIZE];
		char ch = ' ';
		size_t i(0);
		memset(buf, 0, SIZE);

		while ('*' != ch && i < SIZE)
		{
			ch = ERS_Getc(self->portNum_);
			if(ch != EOF)
			{
				buf[i++] = ch;
			}
		}

		/*
		DEBUG_PRINT(buf);
		*/

		std::string s(buf);
		self->processEvent(s);
	}
	_endthreadex(0);
	return NULL;
}

const int Gainer::CONFIG[][4] = 
{
	// N_AIN, N_DIN, N_AOUT, N_DOUT
	{0,},         // dummy
	{4, 4, 4, 4}, // 1
	{8, 0, 4, 4}, // 2
	{4, 4, 8, 0}, // 3
	{8, 0, 8, 0}, // 4
	{0,16, 0, 0}, // 5
	{0, 0, 0,16}, // 6
	{0,}          // 7 for matrix LED
};

